/******************************************************************

    CBDLS.C

    Ismo Krkkinen

    Randomized local search that varies the number of clusters.

******************************************************************/

#define ProgName       "CBDLS"
#define VersionNumber  "Version 0.1.4"
#define LastUpdated    "9.10.2002" /* iak */

#define  FACTFILE  "cbdls.fac"

#include "parametr.c"
#include "dls.h"
#include "cb.h"
#include "file.h"
#include "interfc.h"
#include "memctrl.h"
#include "random.h"
#include "cb_util.h"
#include "DistCrit/distance.h"
#include "DistCrit/criteria.h"
#include "iter.h"
#include "sortcb.h"
#include "logger.h"
#include "dlslog.h"
#include "timer.h"
#include "operatns.h"

#define MAXFILENAME 512

#ifndef max
#define max(a,b) ((a) > (b) ? (a) : (b))
#endif

void PrintInfo() {
    PrintMessage("%s\t%s\n(last modified %s, compiled %s, %s)\n\n"
        "Dynamic local search for varying number of clusters\n"
        "Usage: %s data.ts out.cb [out.pa [in.cb [in.pa]]] %coptions\n\n",
        ProgName, VersionNumber, LastUpdated, __DATE__, __TIME__,
	ProgName, OPTION_SYMBOL);
    PrintOptions();
    PrintMessage("\n");
}

void PrintOperatingInfo() {
    if (Value(QuietLevel)) {
        PrintMessage("\n%s %s %s\n\n", ProgName, VersionNumber, LastUpdated);
        PrintSelectedOptions();
    }
}

static char DataName[MAXFILENAME] = "\0";
static char OutCBName[MAXFILENAME] = "\0";
static char OutPAName[MAXFILENAME] = "\0";
static char InCBName[MAXFILENAME] = "\0";
static char InPAName[MAXFILENAME] = "\0";

#define PARAMETER_FILENAME_COUNT 5

static ParameterInfo paraminfo[PARAMETER_FILENAME_COUNT] = { 
    { DataName, "", 0, INFILE },
    { OutCBName, "", 0, OUTFILE },
    { OutPAName, "", 1, OUTFILE },
    { InCBName, "", 2, INFILE },
    { InPAName, "", 3, INFILE } };

void CheckParameters() {
    if (Value(Distance) == EvaluationFunctionDefault) SetValue(Distance,
	ciCriterionDefaultDistance(Value(EvaluationFunction)));
    if ((Value(HaltingCriterion) == TTimes) && Value(HaltingLimit) < 1001) {
	ErrorMessage(
	    "Multiplier for T-times should be at least 1001 (1.001).\n");
	ExitProcessing(1);
    }
    if ((Value(SwapMethod) == NoSwap) && !Value(GLAIterations)) {
	ErrorMessage("No swap and no GLA-iterations: nothing happens.\n");
	ExitProcessing(2);
    }
    if (ExistFile(OutCBName) && !Value(Overwrite)) {
	ErrorMessage("Output codebook exists and no overwrite specified.\n");
	ExitProcessing(3);
    }
    if (ExistFile(OutPAName) && !Value(Overwrite)) {
	ErrorMessage("Output partition exists and no overwrite specified.\n");
	ExitProcessing(4);
    }
    if (*InCBName && !ExistFile(InCBName)) {
	ErrorMessage("Input codebook not found.\n");
	ExitProcessing(5);
    }
    if (*InPAName && !ExistFile(InPAName)) {
	ErrorMessage("Input partition not found.\n");
	ExitProcessing(6);
    }
    if (!*DataName || !ExistFile(DataName)) {
	ErrorMessage("Input data file not found.\n");
	ExitProcessing(7);
    }
    if (Value(HaltingCriterion) == TTimes) SetValue(HaltingLimit,
	(int)(1000000 / (Value(HaltingLimit) / 1000.0)));
    if (Value(MinNumberOfClusters) > Value(MaxNumberOfClusters))
    {
	ErrorMessage("Max number of clusters less than minimum.\n");
	ExitProcessing(8);
    }
    if (Value(MaxChangeGiven) && Value(GaussGiven)) {
        ErrorMessage("Can't use both max change and Gaussian change.\n");
        ExitProcessing(20);
    }
}

static float MinimumValue = 0;
static float MaximumValue = 0;
static int FileFormat = 0;

static void Initialize(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA) {
    int k;
    float mmin, mmax;
    int foo;
    if (ReadGenericData(DataName, TS,
	&FileFormat, &MinimumValue, &MaximumValue))
    {
	ErrorMessage("Failed to read data.\n");
	ExitProcessing(12);
    }
    if (*InCBName) {
	if (ReadGenericData(InCBName, CB, &foo, &mmin, &mmax)) {
	    ErrorMessage("Failed to read centroids.\n");
	    ExitProcessing(13);
	}
	if (BookSize(CB) > Value(MaxNumberOfClusters)
	    || BookSize(CB) < Value(MinNumberOfClusters))
	{
	    ErrorMessage("Input codebook size %i not in [%i, %i] range.\n",
		BookSize(CB), Value(MinNumberOfClusters),
		Value(MaxNumberOfClusters));
	    ExitProcessing(15);
	}
    } else { /* do random solution */
	/* if no input codebook given, no partitioning either */
	if (FileFormat != RGD_CB && !*OutPAName) {
	    if (Value(QuietLevel) > 1) PrintMessage(
		"Data vectors before duplicate removal: %i,", BookSize(TS));
	    RemoveDuplicatesFromCodebook(TS);
	    if (Value(QuietLevel) > 1)
		PrintMessage(" after: %i.\n", BookSize(TS));
	}
	CreateNewCodebook(CB, Value(MinNumberOfClusters), TS);
	GenerateRandomCodebook(TS, CB);
	VectorFreq(CB, 0) = TotalFreq(TS); /* Since everything goes into C0 */
    }
    if (BookSize(TS) < Value(MaxNumberOfClusters)) {
	ErrorMessage("More clusters than data vectors.\n");
	ExitProcessing(10);
    }
    if (*InPAName) {
	if (ReadGenericMapping(InPAName, TS, PA, &foo)) {
	    ErrorMessage("Failed to read mapping.\n");
	    ExitProcessing(14);
	}
    } else {
	CreateNewPartitioning(PA, TS, BookSize(CB));
	GenerateOptimalPartitioning(TS, CB, PA);
	GenerateOptimalCodebook(TS, CB, PA);
    }
    if (PartitionCount(PA) != BookSize(CB)) {
	ErrorMessage("Partition and codebook sizes do not match.\n");
	ExitProcessing(11);
    }
    for (k = 0; k < BookSize(CB); k++) VectorFreq(CB, k) = CCFreq(PA, k);
}

static void Finalize(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA) {
    if (!*OutPAName) SortCodebook(CB, DATA_ASCENDING);
    else WriteGenericMapping(OutPAName, TS, PA, FileFormat, Value(Overwrite));
    WriteGenericData(OutCBName, CB,
	FileFormat, MinimumValue, MaximumValue, Value(Overwrite));
    FreeCodebook(TS);
    FreeCodebook(CB);
    FreePartitioning(PA);
}


int main(int argc, char** argv) {
    TRAININGSET TS;
    CODEBOOK CB;
    PARTITIONING PA;
    DistanceInfo* DI;
    CriterionInfo* CI;
    Iterations* Iter;
    int ReturnValue;
    float Error;
    Operations* Ops;

    ParseParameters(argc, argv, PARAMETER_FILENAME_COUNT, paraminfo);
    initrandom(Value(RandomNumberSeed));
    CheckParameters();
    PrintOperatingInfo();
    Initialize(&TS, &CB, &PA);
    DI = diNew(&TS, Value(Distance), 0, 0, 0, 0);
    CI = ciNew(&TS, DI, Value(EvaluationFunction), 0, 0, 0, 0, 0, 0);
    Iter = itNew(Value(HaltingCriterion) + 1, Value(HaltingIterations),
	Value(HaltingLimit) / 1000000.0f, 0);
    Ops = opsNew(1e30f,
        Value(MinNumberOfClusters), Value(MaxNumberOfClusters),
        Value(MaxChangeGiven) ? Value(MaxChange) : 0,
        Value(Alpha) / 1000.0f, Value(Adaptivity) / 1000.0f,
        Value(Gauss) / (float) Max(Gauss),
        Value(GroupCountChangeProbability)
            / (float) Max(GroupCountChangeProbability),
        Value(GroupCountChangeMethod), Value(SwapMethod), &TS);
#if defined(USE_LOGGER)
    if (Value(TraceData)) logStream(fopen("log.txt", "wt"));
    else logStream(0);
#else
    logStream(0);
#endif

    if (logOn()) logAdd(DLSLOG_RUN);
    timerSnapshot();
    ReturnValue = DynamicLocalSearch(&TS, &CB, &PA, CI, &Error,
	Value(MaxNumberOfClusters), Ops, Iter,
        Value(GLAIterations), max(0, Value(QuietLevel) - 2));
    timerSnapshot();
    if (logOn()) {
        logAdd1(DLSLOG_WALLTIME,
            (float)timerSeconds + timerMicroseconds / 1000000.0f);
        fclose(logFile);
    }

    if (Value(QuietLevel) && (Value(QuietLevel) < 3))
        PrintMessage("\nError = %.6g\n\n", Error);
    opsDelete(Ops);
    itDelete(Iter);
    ciDelete(CI);
    diDelete(DI);
    Finalize(&TS, &CB, &PA);
    return ReturnValue;
}

