/*-------------------------------------------------------------------

    DLS.C

    3.4.2002

    Ismo Krkkinen

    Dynamic local search.

-------------------------------------------------------------------*/

#define VersionNumber "v0.1.2"

#include <float.h>
#include <stdio.h>
#include <time.h>

#include "dls.h"
#include "cb.h"
#include "interfc.h"
#include "memctrl.h"
#include "random.h"
#include <string.h>
#include "cb_util.h"
#include "logger.h"
#include "dlslog.h"

static void PrintOutput(int Iter, float Error, float Best, int Improved,
    int Output, int Clusters)
{
    switch (Output) {
    case 0: break;
    case DLS_IMPROVED:
	if (Improved) PrintMessage("%i\t%i\t%#.6g\n", Iter, Clusters, Best);
	break;
    default:
	PrintMessage("%i\t%i\t%#9.6g\t%#.6g\n", Iter, Clusters, Best, Error);
	break;
    }
}

static void StoreSolution(TRAININGSET* TS, CODEBOOK* ToCB, PARTITIONING* ToPA,
    CODEBOOK* FromCB, PARTITIONING* FromPA)
{
    if (BookSize(ToCB) < BookSize(FromCB)) {
	IncreaseCodebookSize(ToCB, BookSize(FromCB));
	IncreaseNumberOfPartitions(ToPA, BookSize(FromCB));
    } else if (BookSize(ToCB) > BookSize(FromCB)) {
	DecreaseCodebookSize(ToCB, BookSize(FromCB));
	while (PartitionCount(ToPA) > PartitionCount(FromPA))
            JoinPartitions(TS, ToPA, 0, PartitionCount(ToPA) - 1);
    }
    CopyCodebook(FromCB, ToCB);
    CopyPartitioning(FromPA, ToPA);
}

int DynamicLocalSearch(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA,
    CriterionInfo* CI, float* FinalError, int MaxClusters,
    Operations* Ops, Iterations* Iter, int GLAIterations, int Output)
{
    CODEBOOK CBnew;
    PARTITIONING PAnew;
    int IterationNumber, k, Improved, TotalGLAIterations;
    float Value;
    int ReturnValue = EVERYTHING_OK;
    int *Changes, *Freq2IndexMap;
    if (!strstr(CB->GenerationMethod, "DLS ")) {
	if (*(CB->GenerationMethod)) AddGenerationMethod(CB, " + ");
	AddGenerationMethod(CB, "DLS ");
	AddGenerationMethod(CB, VersionNumber);
    }
    Freq2IndexMap = FrequencyToIndexMap(TS);
    Changes = allocate(sizeof(int) * MaxClusters);
    CreateNewCodebook(&CBnew, BookSize(CB), TS);
    CreateNewPartitioning(&PAnew, TS, BookSize(CB));
    IterationNumber = 0;
    *FinalError = ciEvaluate(CI, CB, PA);
    if (logOn()) {
        logAdd1(DLSLOG_ITER, 0.0f);
        logAdd1(DLSLOG_GROUPCOUNT, (float)BookSize(CB));
        logAdd1(DLSLOG_ERROR, *FinalError);
        logAdd(DLSLOG_REPLACEBEST);
    }
    TotalGLAIterations = 0;
    do {
	++IterationNumber;
	StoreSolution(TS, &CBnew, &PAnew, CB, PA);
        memset(Changes, 0, sizeof(int) * MaxClusters);
	opsPerform(Ops, TS, &CBnew, &PAnew, CI, Changes, Freq2IndexMap);
	Improved = 1;
	for (k = 0; (k < GLAIterations) && Improved; k++) {
	    TotalGLAIterations++;
	    ciCalculateOptimalCentroids(CI, &CBnew, &PAnew, Changes);
	    Improved = ciPartitionOptimally(CI, &CBnew, &PAnew, Changes);
	}
	ciCalculateOptimalCentroids(CI, &CBnew, &PAnew, Changes);
	Value = ciEvaluate(CI, &CBnew, &PAnew);
        if (logOn()) {
            logAdd1(DLSLOG_ITER, (float)IterationNumber);
            logAdd1(DLSLOG_GROUPCOUNT, (float)BookSize(&CBnew));
            logAdd1(DLSLOG_ERROR, Value);
        }
	opsUpdate(Ops, Value);
	Improved = *FinalError > Value;
	if (Improved) {
            if (logOn()) logAdd(DLSLOG_REPLACEBEST);
	    StoreSolution(TS, CB, PA, &CBnew, &PAnew);
	    *FinalError = Value;
	} else if (logOn()) logAdd(DLSLOG_DISCARD);
	if (Output) PrintOutput(IterationNumber, Value, *FinalError,
	    Improved, Output, BookSize(&CBnew));
    } while (!itStop(Iter, *FinalError, BookSize(CB)));
    if (Output) PrintMessage(
	"Average GLA iterations per RLS iteration (max = %i): %f.\n",
	GLAIterations, TotalGLAIterations / (float) (IterationNumber + 1));
    FreePartitioning(&PAnew);
    FreeCodebook(&CBnew);
    deallocate(Changes);
    deallocate(Freq2IndexMap);
    return ReturnValue;
}

