/*
    LRDLS

    Ismo Krkkinen

    Prints desired information from log.txt produced by cbdls.

*/

#define ProgName       "LRDLS"
#define VersionNumber  "Version 0.1.1"
#define LastUpdated    "9.4.2002" /* iak */

#define FACTFILE "lrdls.fac"

extern "C" {

#include "parametr.c"

#include "file.h"
#include "dlslog.h"
#include "logger.h"
#include <stdio.h>
#include "interfc.h"

}

#include <vector>
#include <utility>

#define MAXFILENAME 512

void PrintInfo() {
    PrintMessage("%s\t%s\n(last modified %s, compiled %s, %s)\n\n"
        "Dynamic local search log output analysis program.\n"
        "Usage: %s log.txt %coptions\n\n",
        ProgName, VersionNumber, LastUpdated, __DATE__, __TIME__,
        ProgName, OPTION_SYMBOL);
    PrintOptions();
    PrintMessage("\n");
}

void PrintOperatingInfo() {
    PrintMessage("\n%s %s %s\n\n", ProgName, VersionNumber, LastUpdated);
    PrintSelectedOptions();
}

struct SolutionInfo {
    int M;
    int Iter;
    float Error;
    SolutionInfo(int Groups = 0, int Iterations = 0, float SolutionError = 0.0f)
        : M(Groups), Iter(Iterations), Error(SolutionError) { }
};

struct RunInfo {
    std::vector<SolutionInfo> S;
    float Time;
    RunInfo() : Time(0.0f) { }
};

struct LogInfo {
    std::vector<RunInfo*> R;
};

static char LogName[MAXFILENAME] = "\0";

#define PARAMETER_FILENAME_COUNT 1

static ParameterInfo paraminfo[PARAMETER_FILENAME_COUNT] = { 
    { LogName, "", 0, INFILE }
};

int main(int argc, char** argv) {
    FILE* F;
    ParseParameters(argc, argv, PARAMETER_FILENAME_COUNT, paraminfo);
    F = fopen(LogName, "rt");
    if (!F) {
        ErrorMessage("Failed to open %s for reading.\n\n", LogName);
        return 1;
    }
    PrintOperatingInfo();
    logStream(F);
    LogInfo LI;
    RunInfo* CurrRun = 0;
    SolutionInfo CurrSol;
    // Store log into memory.
    while (!feof(F)) {
        int ReadError;
        float Value;
        const char* ID = logID();
        if (!ID) break;
        if (!strcmp(ID, DLSLOG_RUN)) {
            CurrRun = new RunInfo;
            LI.R.push_back(CurrRun);
        } else if (!strcmp(ID, DLSLOG_WALLTIME)) {
            Value = logValue(&ReadError);
            if (ReadError) {
                ErrorMessage("Error reading %s from log, bad value.\n\n", ID);
                return 2;
            }
            CurrRun->Time = Value;
        } else if (!strcmp(ID, DLSLOG_ITER)) {
            Value = logValue(&ReadError);
            if (ReadError) {
                ErrorMessage("Error reading %s from log, bad value.\n\n", ID);
                return 3;
            }
            CurrSol.Iter = int(Value);
        } else if (!strcmp(ID, DLSLOG_GROUPCOUNT)) {
            Value = logValue(&ReadError);
            if (ReadError) {
                ErrorMessage("Error reading %s from log, bad value.\n\n", ID);
                return 4;
            }
            CurrSol.M = int(Value);
        } else if (!strcmp(ID, DLSLOG_ERROR)) {
            Value = logValue(&ReadError);
            if (ReadError) {
                ErrorMessage("Error reading %s from log, bad value.\n\n", ID);
                return 5;
            }
            CurrSol.Error = Value;
        } else if (!strcmp(ID, DLSLOG_REPLACEBEST)) {
            if (!Value(IterationsLimited)
                || Value(IterationLimit) >= CurrSol.Iter)
                    CurrRun->S.push_back(CurrSol);
        } else if (!strcmp(ID, DLSLOG_DISCARD)) {
            if (!Value(IterationsLimited)
                || Value(IterationLimit) >= CurrSol.Iter)
                    CurrRun->S.push_back(CurrSol);
        } else {
            ErrorMessage("Unknown ID: %s\n\n", ID);
            return 255;
        }
    }
    if (!feof(F)) {
        ErrorMessage("Trouble reading %s\n\n", LogName);
        return 254;
    }
    fclose(F);
    // Analyze data here.
    size_t Run, Sol;
    if (Value(GroupCounts)) {
        int MaxGroups = 0, MinGroups = 0x7fffFFFF;
        for (Run = 0; Run < LI.R.size(); ++Run)
            for (Sol = 0; Sol < LI.R[Run]->S.size(); ++Sol) {
                if (LI.R[Run]->S[Sol].M < MinGroups)
                    MinGroups = LI.R[Run]->S[Sol].M;
                if (LI.R[Run]->S[Sol].M > MaxGroups)
                    MaxGroups = LI.R[Run]->S[Sol].M;
            }
        std::vector<std::pair<int, float> > Table(MaxGroups + 1);
        F = fopen("counts.txt", "wt");
        if (!F) {
            fprintf(stderr, "Failed to open group count file for writing.\n\n");
            return 253;
        }
        std::fill(Table.begin(), Table.end(), make_pair(0,0));
        for (Sol = 0; Sol < Table.size(); ++Sol) Table[Sol].first = Sol;
        float Mean = 0;
        for (Run = 0; Run < LI.R.size(); ++Run) {
            int M = LI.R[Run]->S[0].M;
            float Error = LI.R[Run]->S[0].Error;
            for (Sol = 1; Sol < LI.R[Run]->S.size(); ++Sol)
                if (Error > LI.R[Run]->S[Sol].Error) {
                    M = LI.R[Run]->S[Sol].M;
                    Error = LI.R[Run]->S[Sol].Error;
                }
            ++Table[M].second;
            Mean += M;
        }
        Mean /= LI.R.size();
        if (Value(PrintMeans))
            fprintf(F, "%f\t%i\n", Mean, LI.R.size());
        for (Sol = MinGroups; Sol < Table.size(); ++Sol)
            fprintf(F, "%i\t%f\n", Table[Sol].first, Table[Sol].second);
        fclose(F);
    }
    if (Value(FinalErrors)) {
        F = fopen("errors.txt", "wt");
        if (!F) {
            fprintf(stderr, "Failed to open error value file for writing.\n\n");
            return 252;
        }
        std::vector<float> Table(LI.R.size());
        float Mean = 0;
        for (Run = 0; Run < LI.R.size(); ++Run) {
            float Error = LI.R[Run]->S[0].Error;
            for (Sol = 1; Sol < LI.R[Run]->S.size(); ++Sol)
                if (Error > LI.R[Run]->S[Sol].Error)
                    Error = LI.R[Run]->S[Sol].Error;
            Table[Run] = Error;
            Mean += Error;
        }
        Mean /= LI.R.size();
        if (Value(PrintMeans)) fprintf(F, "%f\n", Mean);
        for (Run = 0; Run < Table.size(); ++Run) fprintf(F, "%f\n", Table[Run]);
        fclose(F);
    }
    if (Value(Iterations)) {
        F = fopen("iters.txt", "wt");
        if (!F) {
            fprintf(stderr, "Failed to open iteration file for writing.\n\n");
            return 251;
        }
        std::vector<int> Table(LI.R.size());
        float Mean = 0;
        for (Run = 0; Run < LI.R.size(); ++Run) {
            Table[Run] = LI.R[Run]->S[LI.R[Run]->S.size() - 1].Iter;
            Mean += Table[Run];
        }
        Mean /= LI.R.size();
        if (Value(PrintMeans)) fprintf(F, "%f\n", Mean);
        for (Run = 0; Run < Table.size(); ++Run) fprintf(F, "%i\n", Table[Run]);
        fclose(F);
    }
    if (Value(WallTimes)) {
        F = fopen("times.txt", "wt");
        if (!F) {
            fprintf(stderr, "Failed to open time file for writing.\n\n");
            return 250;
        }
        std::vector<float> Table(LI.R.size());
        float Mean = 0;
        for (Run = 0; Run < LI.R.size(); ++Run) {
            Table[Run] = LI.R[Run]->Time;
            Mean += Table[Run];
        }
        Mean /= LI.R.size();
        if (Value(PrintMeans)) fprintf(F, "%f\n", Mean);
        for (Run = 0; Run < Table.size(); ++Run) fprintf(F, "%f\n", Table[Run]);
        fclose(F);
    }
    return 0;
}

