/*-------------------------------------------------------------------

    OPERATNS.C

    3.4.2002

    Ismo Krkkinen

    Operation selection for DLS.

-------------------------------------------------------------------*/

#include "operatns.h"
#include "interfc.h"
#include "memctrl.h"
#include "random.h"
#include "cb_util.h"
#include <math.h>


Operations* opsNew(float InitialError, int MinClust, int MaxClust,
    int MaxChg, float ProbDistrPower, float Adaptivity, float GaussStdDev,
    float GroupChangeProb, int CountChangeMethod, int SwapMethod,
    TRAININGSET* TS)
{
    Operations* Ops;
    if (MinClust > MaxClust) {
	ErrorMessage("operatns.c::opsNew: MinClust (%i) > MaxClust (%i).\n",
	    MinClust, MaxClust);
	ExitProcessing(RECOVERABLE_ERROR_UNCONTROLLED_EXIT);
    }
    Ops = allocate(sizeof(Operations));
    Ops->BestError = InitialError;
    Ops->RemoveP = GroupChangeProb / 2;
    Ops->AddP = GroupChangeProb;
    Ops->MinClusters = MinClust;
    Ops->MaxClusters = MaxClust;
    Ops->MaxChange = MaxChg;
    Ops->Alpha = ProbDistrPower;
    Ops->Adaptivity = Adaptivity;
    Ops->Gauss = GaussStdDev;
    Ops->LastSelected = OPS_SEL_NONE;
    Ops->CountChangeMethod = CountChangeMethod;
    Ops->SwapMethod = SwapMethod;
    Ops->Permutation = allocate(sizeof(PermutationData) * MaxClust);
    CreateNewCodebook(&(Ops->Store), MaxClust, TS);
    return Ops;
}

void opsDelete(Operations* Ops) {
    FreeCodebook(&(Ops->Store));
    deallocate(Ops->Permutation);
    deallocate(Ops);
}

int opsSelect(Operations* Ops, int CurrentSize) {
    float R = frand();
    if (R <= Ops->RemoveP && CurrentSize > Ops->MinClusters)
	Ops->LastSelected = OPS_SEL_REMOVE;
    else if (R <= Ops->AddP && CurrentSize < Ops->MaxClusters)
	Ops->LastSelected = OPS_SEL_ADD;
    else Ops->LastSelected = OPS_SEL_SWAP;
    return Ops->LastSelected;
}

void opsUpdate(Operations* Ops, float Error) {
    if (Error < Ops->BestError) {
        Ops->BestError = Error;
        Ops->Alpha *= Ops->Adaptivity;
    }
}

static int perm_asc(const void* a, const void* b) {
    const PermutationData* A = (PermutationData*) a;
    const PermutationData* B = (PermutationData*) b;
    if (A->Key > B->Key) return 1;
    if (A->Key < B->Key) return -1;
    return 0;
}

static int perm_desc(const void* a, const void* b) {
    const PermutationData* A = (PermutationData*) a;
    const PermutationData* B = (PermutationData*) b;
    if (A->Key < B->Key) return 1;
    if (A->Key > B->Key) return -1;
    return 0;
}

static void Permute(PermutationData* Permutation, int Count) {
    int k;
    for (k = 0; k < Count; k++) {
        Permutation[k].Number = k;
        Permutation[k].Key = CombTaus();
    }
    qsort(Permutation, Count, sizeof(PermutationData), &perm_asc);
}

static void SmallestRemove(Operations* Ops, TRAININGSET* TS,
    CODEBOOK* CB, PARTITIONING* PA, int TargetSize)
{
    int k;
    for (k = 0; k < BookSize(CB); ++k) {
        Ops->Permutation[k].Number = k;
        Ops->Permutation[k].Key = CCFreq(PA, k);
    }
    qsort(Ops->Permutation, BookSize(CB), sizeof(PermutationData), &perm_desc);
    for (k = 0; k < TargetSize; k++)
        CopyNode(&Node(CB, Ops->Permutation[k].Number),
            &Node(&(Ops->Store), k), VectorSize(CB));
    DecreaseCodebookSize(CB, TargetSize);
    while (PartitionCount(PA) > TargetSize)
        JoinPartitions(TS, PA, 0, PartitionCount(PA) - 1);
    for (k = 0; k < BookSize(CB); k++)
        CopyNode(&Node(&(Ops->Store), k), &Node(CB, k), VectorSize(CB));
}

/* Ripped from lbgu.c with modifications. */
static int SelectSmall(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA,
    DistanceInfo* DI)
{
    int k, Vect, m;
    float MinError, Error, MyDist, Dist, MinDist;
    int MinIndex;
    MinError = 1e30;
    MinIndex = 0;
    for (k = 0; k < BookSize(CB); ++k) {
        Error = 0;
        Vect = FirstVector(PA, k);
        while (!EndOfPartition(Vect)) {
            MyDist = diDistance(DI, TS, CB, PA, Vect, k, DT_DATA2MODEL);
            if (k)
                MinDist = diDistance(DI, TS, CB, PA, Vect, 0, DT_DATA2MODEL);
            else MinDist = diDistance(DI, TS, CB, PA, Vect, 1, DT_DATA2MODEL);
            for (m = 1; m < BookSize(CB); ++m) if (m != k) {
                Dist = diDistance(DI, TS, CB, PA, Vect, m, DT_DATA2MODEL);
                if (Dist < MinDist) MinDist = Dist;
            }
            Error +=
                VectorFreq(TS, Vect) * (MinDist - MyDist) / CCFreq(PA, k);
            Vect = NextVector(PA, Vect);
        }
        if (Error < MinError) {
            MinError = Error;
            MinIndex = k;
        }
    }
    return MinIndex;
}


static void IterativeRemove(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA,
    int TargetSize, CriterionInfo* CI, int* Changes)
{
    int Smallest;
    while (BookSize(CB) > TargetSize) {
        Smallest = SelectSmall(TS, CB, PA, ciCriterionDistance(CI));
        if (Smallest != BookSize(CB) - 1) {
            CopyNode(&Node(CB, BookSize(CB) - 1), &Node(CB, Smallest),
                VectorSize(CB));
            DecreaseCodebookSize(CB, BookSize(CB) - 1);
            JoinPartitions(TS, PA, Smallest, PartitionCount(PA) - 1);
            Changes[Smallest] = 1;
        } else {
            DecreaseCodebookSize(CB, BookSize(CB) - 1);
            JoinPartitions(TS, PA, 0, PartitionCount(PA) - 1);
            Changes[0] = 1;
        }
        ciPartitionOptimally(CI, CB, PA, Changes);
        ciCalculateOptimalCentroids(CI, CB, PA, Changes);
    }
}

static void RemoveClusters(Operations* Ops,
    TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA,
    int* Changes, CriterionInfo* CI)
{
    int k, Delta, NewSize, MaxChangeCount;
    if (Ops->MaxChange) {
	MaxChangeCount = Ops->MaxChange;
        if (BookSize(CB) - MaxChangeCount < Ops->MinClusters)
            MaxChangeCount = BookSize(CB) - Ops->MinClusters;
        Delta = irand(1, MaxChangeCount);
    } else if (Ops->Gauss) {
	MaxChangeCount = (BookSize(CB) - Ops->MinClusters) - 1;
        do Delta = 1 + (int)(MaxChangeCount * Ops->Gauss * fabs(dgauss()));
        while (Delta > MaxChangeCount + 1);
    } else {
	MaxChangeCount = (BookSize(CB) - Ops->MinClusters) - 1;
	Delta = 1 + (int)(MaxChangeCount * pow(drand(), Ops->Alpha));
	if (Delta > MaxChangeCount + 1) Delta = MaxChangeCount + 1;
    }
    NewSize = BookSize(CB) - Delta;
    if (Ops->CountChangeMethod == OPS_CCM_RANDOM) {
        Permute(Ops->Permutation, BookSize(CB));
        for (k = 0; k < NewSize; k++)
	    CopyNode(&Node(CB, Ops->Permutation[k].Number),
	        &Node(&(Ops->Store), k), VectorSize(CB));
        DecreaseCodebookSize(CB, NewSize);
        while (PartitionCount(PA) > NewSize)
            JoinPartitions(TS, PA, 0, PartitionCount(PA) - 1);
        for (k = 0; k < BookSize(CB); k++)
            CopyNode(&Node(&(Ops->Store), k), &Node(CB, k), VectorSize(CB));
    } else if (Ops->CountChangeMethod == OPS_CCM_SIZE)
        SmallestRemove(Ops, TS, CB, PA, NewSize);
    else IterativeRemove(TS, CB, PA, NewSize, CI, Changes);
    memset(Changes, 1, sizeof(int) * BookSize(CB));
    ciPartitionOptimally(CI, CB, PA, Changes);
    for (k = 0; k < BookSize(CB); k++) VectorFreq(CB, k) = CCFreq(PA, k);
}

static void BiggestAdd(Operations* Ops, TRAININGSET* TS,
    CODEBOOK* CB, PARTITIONING* PA, int TargetSize)
{
    int k, V, Index, m, OriginalBookSize;
    OriginalBookSize = BookSize(CB);
    for (k = 0; k < BookSize(CB); ++k) {
        Ops->Permutation[k].Number = k;
        Ops->Permutation[k].Key = CCFreq(PA, k);
    }
    qsort(Ops->Permutation, BookSize(CB), sizeof(PermutationData), &perm_desc);
    k = 0;
    while (BookSize(CB) < TargetSize) {
        Index = Ops->Permutation[k].Number;
        if (UniqueVectors(PA, Index) > 1) {
            V = FirstVector(PA, Index);
            for (m = 0; m < irand(0, UniqueVectors(PA, Index) - 1); ++m)
                V = NextVector(PA, V);
            if (EqualVectors(Vector(CB, Index), Vector(TS, V), VectorSize(CB)))
            {
                if (V == FirstVector(PA, Index)) V = NextVector(PA, V);
                else V = FirstVector(PA, Index);
            }
            IncreaseCodebookSize(CB, BookSize(CB) + 1);
            CopyVector(Vector(TS, V), Vector(CB, BookSize(CB) - 1),
                VectorSize(CB));
            IncreaseNumberOfPartitions(PA, BookSize(CB));
            ChangePartition(TS, PA, BookSize(CB) - 1, V);
        }
        if (++k == OriginalBookSize) k = 0;
    }
}

/* Ripped with changes from lbgu.c. */
static int SelectBig(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA,
    DistanceInfo* DI)
{
    int k, Vect, m;
    float MaxError, Error, MyDist, Dist, MinDist;
    int MaxIndex;
    MaxError = 0;
    MaxIndex = 0;
    for (k = 0; k < BookSize(CB); ++k) {
        Error = 0;
        Vect = FirstVector(PA, k);
        while (!EndOfPartition(Vect)) {
            MyDist = diDistance(DI, TS, CB, PA, Vect, k, DT_DATA2MODEL);
            if (k)
                MinDist = diDistance(DI, TS, CB, PA, Vect, 0, DT_DATA2MODEL);
            else MinDist = diDistance(DI, TS, CB, PA, Vect, 1, DT_DATA2MODEL);
            for (m = 1; m < BookSize(CB); ++m) if (m != k) {
                Dist = diDistance(DI, TS, CB, PA, Vect, m, DT_DATA2MODEL);
                if (Dist < MinDist) MinDist = Dist;
            }
            Error +=
                VectorFreq(TS, Vect) * (MinDist - MyDist) / CCFreq(PA, k);
            Vect = NextVector(PA, Vect);
        }
        if (Error > MaxError) {
            MaxError = Error;
            MaxIndex = k;
        }
    }
    return MaxIndex;
}

static void IterativeAdd(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA,
    int TargetSize, CriterionInfo* CI, int* Changes)
{
    int Biggest, k, V;
    while (BookSize(CB) < TargetSize) {
        Biggest = SelectBig(TS, CB, PA, ciCriterionDistance(CI));
        V = FirstVector(PA, Biggest);
        if (UniqueVectors(PA, Biggest) > 1)
            for (k = 0; k < irand(0, UniqueVectors(PA, Biggest) - 1); ++k)
                V = NextVector(PA, V);
        if (EqualVectors(
            Vector(CB, Biggest), Vector(TS, V), VectorSize(CB)))
        {
            if (V == FirstVector(PA, Biggest)) V = NextVector(PA, V);
            else V = FirstVector(PA, Biggest);
        }
        IncreaseCodebookSize(CB, BookSize(CB) + 1);
        CopyVector(Vector(TS, V), Vector(CB, BookSize(CB) - 1), VectorSize(CB));
        IncreaseNumberOfPartitions(PA, BookSize(CB));
        ChangePartition(TS, PA, BookSize(CB) - 1, V);
        Changes[BookSize(CB) - 1] = 1;
        ciPartitionOptimally(CI, CB, PA, Changes);
        ciCalculateOptimalCentroids(CI, CB, PA, Changes);
    }
}

static void AddClusters(Operations* Ops,
    TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA,
    int* Changes, CriterionInfo* CI)
{
    int k, n, Delta, MaxChangeCount;
    if (Ops->MaxChange) {
	MaxChangeCount = Ops->MaxChange;
        if (BookSize(CB) + MaxChangeCount > Ops->MaxClusters)
            MaxChangeCount = Ops->MaxClusters - BookSize(CB);
        Delta = irand(1, MaxChangeCount);
    } else if (Ops->Gauss) {
        MaxChangeCount = (Ops->MaxClusters - BookSize(CB)) - 1;
        do Delta = 1 + (int)(MaxChangeCount * Ops->Gauss * fabs(dgauss()));
        while (Delta > MaxChangeCount + 1);
    } else {
	MaxChangeCount = (Ops->MaxClusters - BookSize(CB)) - 1;
	Delta = 1 + (int)(MaxChangeCount * pow(drand(), Ops->Alpha));
	if (Delta > MaxChangeCount + 1) Delta = MaxChangeCount + 1;
    }
    if (Ops->CountChangeMethod == OPS_CCM_RANDOM) {
        IncreaseCodebookSize(CB, BookSize(CB) + Delta);
        IncreaseNumberOfPartitions(PA, BookSize(CB));
        for (k = BookSize(CB) - Delta; k < BookSize(CB); k++) {
            Changes[k] = 1;
            n = SelectNonCodevector(TS, CB, 0);
            CopyVector(Vector(TS, n), Vector(CB, k), VectorSize(CB));
        }
    } else if (Ops->CountChangeMethod == OPS_CCM_SIZE)
        BiggestAdd(Ops, TS, CB, PA, BookSize(CB) + Delta);
    else IterativeAdd(TS, CB, PA, BookSize(CB) + Delta, CI, Changes);
    ciPartitionOptimally(CI, CB, PA, Changes);
    for (k = 0; k < BookSize(CB); k++) VectorFreq(CB, k) = CCFreq(PA, k);
}

void opsPerform(Operations* Ops,
    TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA,
    CriterionInfo* CI, int* Changes, int* Freq2IndexMap)
{
    int From, To;
    switch (opsSelect(Ops, BookSize(CB))) {
    case OPS_SEL_REMOVE:
        RemoveClusters(Ops, TS, CB, PA, Changes, CI);
        break;
    case OPS_SEL_ADD:
        AddClusters(Ops, TS, CB, PA, Changes, CI);
        break;
    case OPS_SEL_SWAP:
        SwapCentroid(CI, TS, CB, PA, Ops->SwapMethod, &From, &To,
            Freq2IndexMap);
        Changes[To] = ciLocalPartition(CI, CB, PA, Changes, To);
        break;
    default:
        ErrorMessage(
            "operatns.c::opsPerform: Unsupported operation.\n");
        ExitProcessing(FATAL_ERROR);
        break;
    }
}

